#!/usr/bin/env python3
from __future__ import absolute_import, division, print_function
import rospy
import numpy as np
from scipy.ndimage import morphology
from nav_msgs.msg import OccupancyGrid, MapMetaData, Path
from geometry_msgs.msg import Pose2D, Pose, PoseStamped, Point, Quaternion
from visualization_msgs.msg import MarkerArray, Marker
from aro_msgs.srv import PlanPath, PlanPathRequest, PlanPathResponse
import tf2_ros
import geometry_msgs.msg

"""
Here are imports that you are most likely will need. However, you may wish to remove or add your own import.
"""


class PathPlanner():

    def __init__(self):
        # Initialize the node
        rospy.init_node("path_planner")

        # Get some useful parameters
        self.mapFrame = rospy.get_param("~map_frame", "map")
        self.robotFrame = rospy.get_param("~robot_frame", "base_link")
        self.robotDiameter = float(rospy.get_param("~robot_diameter", 0.6))
        self.occupancyThreshold = int(rospy.get_param("~occupancy_threshold", 25))

        # Helper variable to determine if grid was received at least once
        self.gridReady = False

        # You may wish to listen to the transformations of the robot
        self.tfBuffer = tf2_ros.Buffer()
        # Use the tfBuffer to obtain transformation as needed
        listener = tf2_ros.TransformListener(self.tfBuffer)

        # Subscribe to grid
        self.gridSubscriber = rospy.Subscriber('occupancy', OccupancyGrid, self.grid_cb)

        # Publishers for visualization
        self.path_vis_pub = rospy.Publisher('path', Path, queue_size=1)
        self.start_and_goal_vis_pub = rospy.Publisher('start_and_goal', MarkerArray, queue_size=1)

        rospy.loginfo('Path planner initialized.')
    
    # nav_msgs.msg.Path can be visualized in RViz
    def publishPath(self, path_2d):
        msg = Path() 
        msg.header.frame_id = self.mapFrame
        msg.header.stamp = rospy.get_rostime()
        for waypoint in path_2d:
            pose = PoseStamped()
            pose.header.frame_id = self.mapFrame
            pose.pose.position.x = waypoint.x
            pose.pose.position.y = waypoint.y 
            pose.pose.position.z = 0
            msg.poses.append(pose)
        
        rospy.loginfo("Publishing plan.")
        self.path_vis_pub.publish(msg) 

    # MarkerArray can be visualized in RViz
    def publishStartAndGoal(self, start, goal):
        msg = MarkerArray() 
        m_start = Marker()
        m_start.header.frame_id = self.mapFrame
        m_start.id = 1
        m_start.type = 2
        m_start.action = 0
        m_start.pose = Pose()
        m_start.pose.orientation = Quaternion(0.0, 0.0, 0.0, 1.0)
        m_start.pose.position = Point(start.x, start.y, 0.0) 
        # m_start.points.append(Point(start.x, start.y, 0.0))
        m_start.color.r = 1.0
        m_start.color.g = 0.0
        m_start.color.b = 0.0
        m_start.color.a = 0.8
        m_start.scale.x = 0.1
        m_start.scale.y = 0.1
        m_start.scale.z = 0.001
        msg.markers.append(m_start)

        # goal marker
        m_goal = Marker()
        m_goal.header.frame_id = self.mapFrame
        m_goal.id = 2
        m_goal.type = 2
        m_goal.action = 0
        m_goal.pose = Pose()
        m_goal.pose.orientation = Quaternion(0.0, 0.0, 0.0, 1.0)
        m_goal.pose.position = Point(goal.x, goal.y, 0.0) 
        # m_start.points.append(Point(start.x, start.y, 0.0))
        m_goal.color.r = 0.0
        m_goal.color.g = 1.0
        m_goal.color.b = 0.0
        m_goal.color.a = 0.8
        m_goal.scale.x = 0.1
        m_goal.scale.y = 0.1
        m_goal.scale.z = 0.001
        msg.markers.append(m_goal)
        rospy.loginfo("Publishing start and goal markers.")
        self.start_and_goal_vis_pub.publish(msg) 

    def planPath(self, request):
        """ Plan and return path from the start to the requested goal """

        # TODO: Get the position of the start and goal (real-world)
        startPosition = getGridPosition(request.start, self.gridInfo)
        goalPosition = getGridPosition(request.goal, self.gridInfo)

        self.publishStartAndGoal(request.start, request.goal)

        # TODO: Copy the occupancy grid into some temporary variable and inflate the obstacles depending on robot size.
        # Make sure you take into accout unknown grid tiles as non-traversable and also inflate those.

        # TODO: Compute the path, i.e. run some graph-based search algorithm.
        path = []

        real_path = [Pose2D(pos[0], pos[1], 0) for pos in
                     [""" transform each point into real-world coordinates """ for waypoint in path]]
        response = PlanPathResponse(real_path)

        # Publish planned path for visualization in Rviz.
        self.publishPath(response.path)

        return response

    def extractGrid(self, msg):
        # TODO: extract grid from msg.data and other usefull information
        pass

    def grid_cb(self, msg):
        self.extractGrid(msg)
        if not self.gridReady:
            # Create services
            self.plan_service = rospy.Service('plan_path', PlanPath, self.planPath)
            self.gridReady = True


if __name__ == "__main__":
    pp = PathPlanner()
    rospy.spin()
